/*
**  kdialppp.cpp
**
**  Development History:
**      who                  when           why
**      muquit@muquit.com   Jul-25-2001     first cut
*/

#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <unistd.h> /* fork */
#include <time.h>
#include <stdarg.h>

#include <qcheckbox.h>
#include <qdialog.h>
#include <qframe.h>
#include <qlabel.h>
#include <qlayout.h>
#include <qlined.h>
#include <qmenudata.h>
#include <qpainter.h>
#include <qpopupmenu.h>
#include <qpushbutton.h>
#include <qspinbox.h>
#include <qtooltip.h>
#include <qfile.h>

#include <kconfig.h>
#include <kiconloader.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <knotifyclient.h>
#include <kwin.h>
#include <krun.h>
#include <kurllabel.h>

#include <kdialogbase.h>
#include "kdialppp.h"


#include <net/if.h>
#include <net/ppp_defs.h>
#include <net/if_ppp.h>

#include "pixmaps/red.xpm"
#include "pixmaps/green.xpm"
#include "pixmaps/gdot.xpm"
#include "pixmaps/rdot.xpm"

//#undef DEBUG

static bool high_timer_started=false;
static int pid=(-1);
static bool stopped=false;

extern "C"
{
/*
**  processEnded()
**      routine which handles SIGCHLD
**
**  Parameters:
**      int signum 
**
**  Return Values:
**      none
**
**  Limitations and Comments:
**      called when a child dies. the parent make sure it doesn't become a
**      zombie. uses statics: pid, high_timer_started and stopped
**
**  Development History:
**      who                  when           why
**      muquit@muquit.com   Aug-09-2001     first cut
*/
static void processEnded(int signum)
{
    int
        lpid;

    char
        *code;

    lpid=waitpid(-1,(int *) 0,WNOHANG);
    if (lpid == -1)
        return;

#ifdef DEBUG
    (void) fprintf(stderr,"processEnded() - process %d ended\n",lpid);
#endif /* DEBUG */
    pid=(-1);
    high_timer_started=false;
    stopped=true;
}

/*
** return time stamp
** borrowed from Apache long long time ago
*/
char *getTime()
{
    time_t
        t;

    char
        *time_string;

    t=time(NULL);
    time_string=ctime(&t);
    time_string[strlen(time_string)-1]='\0';

    return (time_string);
}
} /* extern "C" */


/*
**  Kdialppp()
**      the constructor
**
**  Development History:
**      who                  when           why
**      muquit@muquit.com   Aug-09-2001     first cut
*/
Kdialppp::Kdialppp() : KSystemTray()
{
    struct sigaction
        act;

    KGlobal::locale()->insertCatalogue("kdialppp");
    conf=kapp->config();
    conf->setGroup("General");

    setBackgroundMode(X11ParentRelative);
    red_ball=new QPixmap(red_xpm);
    green_ball=new QPixmap(green_xpm);
    green_dot=new QPixmap(gdot_xpm);
    red_dot=new QPixmap(rdot_xpm);

    // create popup menu
    popup_menu=new QPopupMenu();

    popup_menu->insertItem(i18n("&About"),
                this,
                SLOT(showAbout()));

    popup_menu->insertItem(SmallIcon("configure"),
                i18n("&Configure..."),
                this,
                SLOT(conFig()));

    popup_menu->insertItem(i18n("&Disconnect"),
                this,
                SLOT(disConnect()));

    popup_menu->insertItem(i18n("&Stats"),
                this,
                SLOT(showStats()));

    popup_menu->insertSeparator();

    popup_menu->insertItem(SmallIcon("exit"),
                i18n("Quit"),
                this,
                SLOT(doQuit()));

    dialer_command=conf->readEntry("dialer_command");
    myDebug("dialer command=%s",dialer_command.latin1());

    // setup the signal handler for child death detection
    sigemptyset(&act.sa_mask);
    act.sa_flags=0;
    act.sa_handler=processEnded;
    sigaction(SIGCHLD,&act,NULL);

    pid=(-1);
    ppp_up=false;
    myDebug("in constructior");
    ppp_state= !ppp_up;
    running=false;
    ready=false;
    timer_started=false;

    last_in_draw=0;
    last_out_draw=0;
    last_in_bytes=0;
    in_bytes=0;
    last_out_bytes=0;
    out_bytes=0;
    painted_yet=false;
    starting=false;
    high_timer_started=false;
    last_starting_draw=0;
    reset_timer=true;

    stat_dlg=0;

    start();
}


/*
**  ~Kdialppp()
**      the destructor
**
**  Development History:
**      who                  when           why
**      muquit@muquit.com   Aug-09-2001     first cut
*/
Kdialppp::~Kdialppp()
{
    delete red_ball;
    delete green_ball;
    delete red_dot;
    delete green_dot;
}


/*
**  mousePressEvent()
**      handles mouse press event.
**
**  Parameters:
**      QMouseEvent *event
**
**  Return Values:
**      none
**
**  Limitations and Comments:
**      start dialer if left mouse is clicked, popup menu if right mouse
**      is clicked.
**
**  Development History:
**      who                  when           why
**      muquit@muquit.com   Aug-09-2001     first cut
*/
void Kdialppp::mousePressEvent(QMouseEvent *event)
{
    char
        buf[BUFSIZ],
        errbuf[BUFSIZ];

    if (event->button() == RightButton)
    {
        myDebug("Right button pressed");
        popup_menu->popup(QCursor::pos());
    }

    if (event->button() == LeftButton)
    {
        myDebug("starting ppp");
        (void) strcpy(buf,PPP_COMMAND);
        if (pid < 0)
        {
            // start with hight interval for low refresh rate
            if (ppp_up == false && dialer_command.isEmpty())
            {
                KMessageBox::error(this,i18n("Please configure your dialer command first"));
                return;
            }
            starting=true;
            stopped=false;
            int rc=execCommand((char *)dialer_command.latin1(),
                    errbuf,strlen(buf)-1);
            myDebug("rc from execCommand()=%d",rc);
        }
        else
        {
            myDebug("Already running pid=%d",pid);
        }
    }

}

// not used
void Kdialppp::enterEvent(QEvent *e)
{
    QLabel::enterEvent(e);
    myDebug("mouse entered");
}

// not used
void Kdialppp::leaveEvent(QEvent *e)
{
    QLabel::enterEvent(e);
    myDebug("mouse left");
}



/*
**  updatePPPInfo()
**
**  Limitations and Comments:
**      open the ppp device /dev/ppp0 and determines received and sent
**      bytes. sets member vars accordingly
**
**  Development History:
**      who                  when           why
**      muquit@muquit.com   Aug-09-2001     first cut
*/
void Kdialppp::updatePPPInfo()
{
    static int
        ifd=(-1);

    struct ifreq
        ifr;

    struct ifpppstatsreq
        req;

    int
        rc;

    static char
        *ppp_device="ppp0"; /* XXX */


    in_bytes=0;
    out_bytes=0;

    ppp_state=ppp_up;

    if (ifd == -1)
    {
        ifd=socket(AF_INET,SOCK_DGRAM,0);
        if (ifd <= 0)
        {
            perror("socket:"); /* XXX */
            ppp_up=false;
            return;
        }
    }
    memset(&ifr,0,sizeof(ifr));
    (void) strcpy(ifr.ifr_name,ppp_device);

    rc=ioctl(ifd,SIOCGIFFLAGS,(char *) &ifr);
    if (rc < 0)
    {
        /* perror("ioctl SIOCGIFFLAGS"); */
        ppp_up=false;
        return;
    }
    memset(&req,0,sizeof(req));
    req.stats_ptr = (caddr_t) &req.stats;

    (void) strcpy(req.ifr__name,ppp_device);
    rc=ioctl(ifd,SIOCGPPPSTATS,&req);
    if (rc < 0)
    {
        /* perror("ioctl SIOCGPPPSTATS"); */
        ppp_up=false;
        return;
    }
    in_bytes=(unsigned long)req.stats.p.ppp_ibytes;
    out_bytes=(unsigned long)req.stats.p.ppp_obytes;

    if (in_bytes == 0 && out_bytes == 0)
        ppp_up=false;
    else
        ppp_up=true;
}


//start the monitor timer
void Kdialppp::start()
{

    myDebug("in start() starting timer (500)");
    timer_id=startTimer(500);
    repaint();
}


// timer event is handled
void Kdialppp::timerEvent(QTimerEvent *)
{
    last_in_bytes=in_bytes;
    last_out_bytes=out_bytes;
    updatePPPInfo();
    if (ppp_up == true)
    {
        if (stat_dlg && stat_dlg->isVisible())
        {
            myDebug("Update stats");
        }
    }

//    myDebug("timerEvent() ppp_up=%d",ppp_up);
//    myDebug("timerEvent() RD=%d SD=%d",in_bytes,out_bytes);

    // if ppp is gettting started, just repaing, we will stop the low res
    // timer and start the high res timer when the link is up
    if (starting == true && ppp_up == false)
    {
        last_starting_draw = !last_starting_draw;
        high_timer_started=false;
        repaint();
        return;
    }

    // link is up, start the high res timer
    if (starting == false && ppp_up == true)
    {
        if (high_timer_started == false)
        {
            myDebug("timerEvent() killing slow timer(500)");
            killTimer(timer_id);
            myDebug("timerEvent() starting fast timer(100)");
            timer_id=startTimer(100);
            high_timer_started=true;
            reset_timer=true;
            stopped=false;
        }
    }
    if (stopped == true)
    {
        myDebug("timerEvent() ppp stopped");
        ppp_up=false;
        starting=false;
        killTimer(timer_id);
        myDebug("resetting timer to 500");
        timer_id=startTimer(500);
        repaint();
        return;
    }

    if (ppp_up == false && reset_timer == true)
    {
        killTimer(timer_id);
        myDebug("timerEvent() staring slow timer(500) again\n");
        timer_id=startTimer(500);
        reset_timer=false;
    }


    if (ppp_state != ppp_up)
    {
        myDebug("timerEvent() calling repaint()");
        starting=false;
        repaint();
    }

    if (last_in_bytes != in_bytes)
    {
        painted_yet=false;
        last_in_draw = !last_in_draw;
        repaint();
    }

    if (last_out_bytes != out_bytes)
    {
        painted_yet=false;
        last_out_draw = !last_out_draw;
        repaint();
    }
    if ((last_in_bytes == in_bytes) && (last_out_bytes == out_bytes))
    {
        if (!painted_yet)
            repaint();
        painted_yet=true;
    }

}

// returns 0 if the process is still running, -1 otherwise
int Kdialppp::isStillRunning(void)
{
    if (pid > 0)
    {
        int
            rc=kill(pid,0);
        if (rc < 0)
            rc=(-1);
        return (rc);
    }
    return (-1);
}

void Kdialppp::disConnect(void)
{
    if (pid > 0)
    {
        myDebug("disConnect() sending KILL signal to pid: %d",pid);

        int 
            rc=kill(pid,SIGKILL);
        myDebug("disConnect() rc from kill=%d",rc);
        if (rc == 0)
            pid=(-1);
    }
    else
        myDebug("disConnect() nothing to kill, pid=%d",pid);
}

/*
**  parseArgs()
**      split commands to NULL terminated individual args
**
**  Parameters:
**      char    *buf     - the command
**      char    **args   - NULL terminated args, returns
**      int     args_len - number of pointer args can hold
**      int     *n       - number of args in the command, returns
**
**  Return Values:
**      0   on success
**      -1 on failure
**
**  Limitations and Comments:
**      command can have 128 args max. This routine is used in Unix only.
**
**  Development History:
**      who                  when           why
**      muquit@muquit.com   May-17-2001     first cut for fltk fluid
**      muquit@muquit.com   Jul-28-2001     for kdialppp
*/
int Kdialppp::parseArgs(char *buf,char **args,int args_len,int *n)
{
    *n=0;
    while (*buf != '\0')
    {
        while ((*buf == ' ')  ||
               (*buf == '\t') ||
               (*buf == '\n'))
        {
            *buf++='\0';
        }
        if (++(*n) > args_len)
        {
            (void) fprintf(stderr,"args buffer overflowed: %d > % d\n",
                           *n,args_len);
            return(-1);
        }
        *args++=buf;

        /* skip over the argument */
        while ((*buf != '\0') &&
               (*buf != ' ')  &&
               (*buf != '\t') &&
               (*buf != '\n'))
            buf++;
    }
    *args='\0';
    return (0);
}

/*
**  execCommand()
**      executes a external command
**
**  Parameters:
**      char *command   - the command to execute
**      char *errbuf    - returns error message if any
**      int  buflen     - length of errbuf
**
**  Return Values:
**      0   on success
**      -1  on failure
**
**  Limitations and Comments:
**      For Unix and Win32. For Win32, it starts a new thread and start
**      the editor from the thread handler routine and the handler routine
**      waits till the editor is done (In Unix, the signal handler does
**      that).
**
**  Development History:
**      who                  when           why
**      muquit@muquit.com    May-15-2001    first cut for fltk fluid
**      muquit@muquit.com    Jul-28-2001    modified for kdialppp. I'm
**                                          leaving the WIN32 code anyways.
*/
int Kdialppp::execCommand(char *command,char *errbuf,int buflen)
{
    int
        rc,
        n=0;

    char
        *args[128],
        cmd_buf[BUFSIZ];

#ifdef WIN32
    HANDLE
        thread_handle;

    DWORD
        ret_val;

    unsigned int
        thread_id;
#endif /* WIN32 */

    (void) strcpy(cmd_buf,command);
#ifndef WIN32 /* UNIX */
    rc=parseArgs(cmd_buf,args,127,&n);
    if (rc != 0)
    {
        (void) fprintf(stderr,"Could not parse:\"%s\"\n",command);
        (void) sprintf(errbuf,"Could not parse:\"%s\"",command);
        KMessageBox::error(this,i18n(errbuf));
        return (-1);
    }

#ifdef DEBUG
    if (n)
    {
        for (int i=0; i < n; i++)
        {
            myDebug("execCommand() - args[%d]=%s",i,args[i]);
        }
    }
#endif /* DEBUG */
#endif /* UNIX */
#ifdef WIN32
    /* start a thread */
    thread_handle=(HANDLE) _beginthreadex(NULL,
            0,
            editorThread,
            (void *) cmd_buf,
            0,
            &thread_id);
#else /* UNIX */

    pid=fork();
    switch(pid)
    {
        case -1:
        {
            (void) strcpy(errbuf,"fork() failed out of process\r\n");
            perror("fork");
            return(-1);
            break;

        }
        case 0: /* child */
        {
            execvp(args[0],args);
            (void) sprintf(errbuf,"failed to exec: %s",command);
            (void) fprintf(stderr,"%s\n",errbuf);
            perror("execvp");
            myDebug("command=%s",dialer_command.latin1());
            myDebug("execCommand() could not exec: %s",command);
            _exit(255);
            break;
        }

        default:
        {
            break;
        }
    }

#endif /* ! WIN32 (UNIX) */

    return(0);
}


void Kdialppp::doQuit(void)
{
    disConnect();
    exit(0);
}

void Kdialppp::homePage(const QString& url)
{
    (void) new KRun(url);
}

void Kdialppp::mailTo(const QString& url)
{
    (void) new KRun(url);
}

// show the about dialog
void Kdialppp::showAbout(void)
{
    static KDialogBase
        *aboutd;
    
    if (!aboutd)
    {
        aboutd=new KDialogBase(KDialogBase::Plain,
                    i18n("About kdialppp"),
                    KDialogBase::Ok,
                    KDialogBase::Ok,
                    this,"about",
                    false);

        QFont logo_font=QFont::defaultFont();
        logo_font.setPointSize(logo_font.pointSize() * 1.5);
        logo_font.setBold(true);

        KURLLabel *logo_label=new KURLLabel(aboutd->plainPage());
        logo_label->setURL("http://www.muquit.com/muquit/software/");
        logo_label->setFont(logo_font);
        logo_label->setText("Kdialppp");
        logo_label->setUnderline(false);
        logo_label->setFloat(true);
        connect(logo_label,SIGNAL(leftClickedURL(const QString&)),
                SLOT(homePage(const QString&)));

        QLabel *version_label=new QLabel(aboutd->plainPage());
        version_label->setText(QString("\nVersion %1\nCopyright (c) 2001 Muhammad A Muquit").arg("1.1"));

        KURLLabel *email_label=new KURLLabel(aboutd->plainPage());
        email_label->setText("muquit@muquit.com");
        email_label->setURL("mailto:muquit@muquit.com");
        email_label->setUnderline(false);
        email_label->setGlow(false);
        email_label->setFloat(true);
        connect(email_label,SIGNAL(leftClickedURL(const QString&)),
                SLOT(mailTo(const QString&)));

        QVBoxLayout *text_layout = new QVBoxLayout(0);
        text_layout->addWidget(version_label);
        text_layout->addWidget(email_label);

        QGridLayout *about_layout = new QGridLayout(aboutd->plainPage(),2,1,12,0);
        about_layout->addWidget(logo_label,0,0,0);
        about_layout->addLayout(text_layout,1,0);
        about_layout->setColStretch(0,2);
    }
    aboutd->show();
}

void Kdialppp::paintEvent(QPaintEvent *e)
{
    QPixmap
        *rd_pix,
        *gd_pix,
        *pix2,
        *pix3,
        *pix;

    // draw background
    //QWidget::paintEvent(e);


    static bool
        which_pix_in=0,
        which_pix_out=0;

//   myDebug("paintEvent()");

    if (ppp_up == false)
        pix=red_ball;
    else
        pix=green_ball;

    pix2=green_dot;
    pix3=green_dot;

    gd_pix=green_dot;
    rd_pix=red_dot;

    int
        x=1 + (12 - pix->width()/2);

    int
        y=1 + (12 - pix->height()/2);


    QPainter p(this);

//    myDebug("Drawing main pixmap x,y=%d,%d",x,y);

    p.drawPixmap(x,y,*pix);
    
    // blink red & green for showing ppp is getting started
    if (ppp_up == false && starting == true && stopped == false)
    {
        myDebug("starting...");
        if (last_starting_draw)
            pix=gd_pix;
        else
            pix=rd_pix;
        p.drawPixmap(6,9,*pix);
        p.drawPixmap(14,9,*pix);
    }

    if (last_in_bytes != in_bytes)
    {
        if (last_in_draw)
            pix2=green_dot;
        else
            pix2=red_dot;
        p.drawPixmap(6,9,*pix2);
        p.drawPixmap(14,9,*pix3);
    }

    if (last_out_bytes != out_bytes)
    {
        if (last_out_draw)
            pix3=green_dot;
        else
            pix3=red_dot;
        p.drawPixmap(6,9,*pix2);
        p.drawPixmap(14,9,*pix3);
    } 

    if ((last_in_bytes == in_bytes) && (last_out_bytes == out_bytes) &&
            ppp_up == true)
    {
        p.drawPixmap(6,9,*gd_pix);
        p.drawPixmap(14,9,*gd_pix);
    }
    p.end();
}

void Kdialppp::myDebug(const char *format,...)
{
#ifdef DEBUG
    va_list
        args;

    va_start(args,format);
    (void) fprintf(stderr,"%s - ",getTime());
    vfprintf(stderr,format,args);
    (void)fprintf(stderr,"\n");
    va_end(args);
#endif /* DEBUG */
}

// Configure dialog
void Kdialppp::conFig(void)
{
    KDialogBase
        *dlg=new KDialogBase(KDialogBase::Plain,
                i18n("Configure Kdialppp"),
                KDialogBase::Ok|KDialogBase::Cancel|KDialogBase::Help,
                KDialogBase::Ok,
                this,
                "config",
                true);

    QVBoxLayout *box=new QVBoxLayout(dlg->plainPage());
    QGridLayout *grid=new QGridLayout(2,2); // 2 row, 2 col
    box->addLayout(grid);

    QLineEdit *command=new QLineEdit(dlg->plainPage());
    command->setText(i18n(PPP_COMMAND));
    QLabel *l=new QLabel(command,
                i18n("Dialer command:"),
                dlg->plainPage());
    grid->addWidget(l,0,0);
    grid->addWidget(command,0,1);

    if (dlg->exec() == QDialog::Accepted)
    {
        myDebug("OK");
        dialer_command=command->text();
        myDebug("dialer=%s",dialer_command.latin1());
        conf=kapp->config();
        conf->setGroup("General");
        conf->writeEntry("dialer_command",dialer_command);
        conf->sync();
    }
}

// show received and sent bytes in a dialog
void Kdialppp::showStats(void)
{
    stat_dlg=new KDialogBase(KDialogBase::Plain,
            i18n("Configure Kdialppp"),
            KDialogBase::Ok|KDialogBase::Cancel|KDialogBase::Help,
            KDialogBase::Ok,
            this,
            "config",
            false);

    stat_dlg->show();
}
